BACKBONE_EMBEDDINGS_KEY = "BACKBONE//EMBEDDINGS"
BACKBONE_TOKENS_KEY = "BACKBONE//TOKENS"


MODEL_PREFIX = "MODEL//"
MODEL_EMBEDDINGS_KEY = "MODEL//EMBEDDINGS"
MODEL_TOKENS_KEY = "MODEL//TOKENS"
MODEL_BATCH_LOSS_KEY = "MODEL//BATCH_LOSS"
MODEL_LOSS_KEY = "MODEL//LOSS"
MODEL_LOGITS_KEY = "MODEL//LOGITS"
MODEL_LOGITS_SEQUENCE_KEY = "MODEL//LOGITS_SEQUENCE"
MODEL_PRED_PROBA_KEY = "MODEL//PRED_PROBA"
MODEL_PRED_STATUS_KEY = "MODEL//PRED_STATUS"


GENERATE_PREFIX = "GENERATE//"
ACTUAL_FUTURE = "ACTUAL_FUTURE"
INPUT_DATA = "INPUT_DATA"
